package Vargus::Common;

#    Copyright (C) 2010-2012 Michael A. Kangin <mak@complife.ru>
#
#    Copyright: Vargus is under GNU GPL, the GNU General Public License.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; version 2 dated June, 1991.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    http://www.gnu.org/licenses/gpl-2.0.html
#


use strict;
#use warnings;

use Sys::Syslog qw(openlog syslog closelog LOG_DAEMON LOG_ERR LOG_WARNING LOG_NOTICE LOG_INFO LOG_DEBUG) ;
use Net::Telnet;
use Time::HiRes qw(sleep gettimeofday);
use Socket;
use Sys::Hostname;

use base "Exporter";
our @EXPORT = qw(
	s_log
	log_n_die
	do_remote_query
	expand_macroses
	find_port_by_name
	cleance_config
	get_cfg_section
	delete_cfg_section
	get_cfg_section_as_string
	get_option
	do_telnet_exchange
	);

sub s_log {
	my $log_priority = shift;
	my $log_message = shift;
	my $date_label;
	my $daemon_role = $main::daemon_role;
	my $daemon_pid = $main::daemon_pid;
	my $daemon_mode = defined($main::daemon_mode) ? $main::daemon_mode : 0;
	my $daemon_name = $0;
	$daemon_name =~ s|.*/||;
	$daemon_name =~ s| .*||;
	$daemon_name = $main::daemon_name if defined($main::daemon_name);

	my ($now_sec, $now_usec) = gettimeofday();
	$date_label = sprintf("%04d-%02d-%02d %02d:%02d:%02d", 
		sub {($_[5]+1900, $_[4]+1, $_[3], $_[2], $_[1], $_[0])}->(localtime($now_sec)))
		unless $daemon_mode;
	$date_label .= sprintf(".%03d", $now_usec);
	openlog($date_label . " $daemon_name($daemon_role/$daemon_pid)", "ndelay,pid,perror", LOG_DAEMON);
	syslog($log_priority, $log_message);
	closelog;
}

sub log_n_die {
	s_log(LOG_ERR, shift);
	&main::prepare_to_die(1);
}


sub get_free_port {
	my $proto = shift;
	my $port_base = shift;
	my $port_name = shift;

	my @engaged_ports;
	my $port;
	my $stat_file;

	$proto = getprotobyname($proto) or log_n_die("Bad proto: $!");
	my $sock_type = $proto==17?SOCK_DGRAM:SOCK_STREAM;

	foreach $stat_file (glob($main::stat_dir . '/*.stat')) {
		open(STAT_FILE, $stat_file) or log_n_die "Error open $stat_file";
		push(@engaged_ports, <STAT_FILE>);
		close(STAT_FILE);
	}

	chomp(@engaged_ports);

	socket(SOCKET, PF_INET, $sock_type, $proto) || log_n_die "Get socket problem: $!";

	if ($port = (grep(/^r:$port_name:/, @engaged_ports))[0]) {
		$port =~ s/.*://;
		if ($proto == 6) {
			bind(SOCKET, sockaddr_in($port, INADDR_ANY)) or $port = 0;
		}
		s_log(LOG_WARNING, "Assign desired by name ($port_name) port $port") if $port;
	}
		
	if (! $port) {
		($port) = $port_base =~ /^(\d+)$/ or log_n_die "Invalid port base $port_base";
		$port++ 
			until (!grep(/:$port$/, @engaged_ports) && 
				bind(SOCKET, sockaddr_in($port, INADDR_ANY))) || 
			$port >= $port_base + 1000;

		log_n_die("Can't find free port from $port_base to " . $port) 
			if $port >= $port_base + 1000;
	}

	close(SOCKET);
	return $port;
}

sub find_port_by_name {
	my $port_name = shift;

	my @engaged_ports;
	my $port;

	foreach my $stat_file (glob($main::stat_dir . '/*.stat')) {
		open(STAT_FILE, $stat_file) or log_n_die "Error open $stat_file";
		push(@engaged_ports, <STAT_FILE>);
		close(STAT_FILE);
	}

	chomp(@engaged_ports);
	
	$port = (grep(/^w:$port_name:/, @engaged_ports))[0];
	$port =~ s/.*://;
	return $port;
}

sub do_remote_query {
	my $host = shift;
	my $query = shift;
	my $answer;
	my $inquiring = new Net::Telnet (
				Timeout => 2,
				errmode => "return",
			);
	my $telnet_status = $inquiring->open( Host => $host, Port => $main::informer_port ) ;
	unless ($telnet_status) {
		s_log(LOG_WARNING, "Remote query error: connection failed ($host:$main::informer_port)");
		return;
	}

	$inquiring->print($query);
	$inquiring->print("exit");
	while (my $responce = $inquiring->getline()) {
		chomp($responce);
		next if $responce =~ /^(:|>)/;
		$answer .= "$responce\n";
	}

	$inquiring->close();
	return $answer;
}


sub find_remote_port {
	my $host = shift;
	my $port_name = shift;

	my $port = do_remote_query($host, "findport $port_name");
	chomp($port);
	unless ($port) {
		s_log(LOG_NOTICE, "Can't find remote port $port_name for host $host");
		return;
	}
	return $port;
}



sub expand_macroses {
	my $macro;
	my $macro_string;
	my @unmacro_strings;
	my $conf_dir = $main::conf_dir;

	my @macro_array = @_;
	foreach $macro_string (@macro_array) {

	while ($macro_string =~ /\(.*\)/) {
		$macro = $macro_string;
		$macro =~ s/.*?\(([^\(]*?)\).*/$1/;
		if ($macro eq "date") { 
			$macro = sprintf("%04d-%02d-%02d", 
				sub {($_[5]+1900, $_[4]+1, $_[3])}->(localtime)); 
		}
		if ($macro eq "time") {  
			$macro = sprintf("%02d-%02d-%02d",
				sub {($_[2], $_[1], $_[0])}->(localtime)); 
		}
		if ($macro =~/^date_of:/) {
			my $custom_time = (split(':', $macro, 2))[1];
			$macro = sprintf("%04d-%02d-%02d", 
				sub {($_[5]+1900, $_[4]+1, $_[3])}->(localtime($custom_time))); 
		}
		if ($macro =~/^time_of:/) {
			my $custom_time = (split(':', $macro, 2))[1];
			$macro = sprintf("%02d-%02d-%02d",
				sub {($_[2], $_[1], $_[0])}->(localtime($custom_time))); 
		}
		if ($macro eq "space") {
			$macro = ' ';
		}
		if ($macro =~ /^include:/) {
			my $include_file = $conf_dir . '/' . (split(':', $macro))[1];
			-r $include_file or log_n_die("Can't read include file $include_file");
			open(INC, $include_file);
			chomp (my @include_body = <INC>);
			close(INC);
			@include_body=cleance_config(@include_body);
			$macro = join('', @include_body);
			$macro =~ tr/ \t\n//d;
		}
		if ($macro =~ /^get_port:/) {
			(undef, my $port_name, my $proto, my $port_base) = split(':', $macro);

			if ($main::generated_ports{$port_name}) {
				$macro = $main::generated_ports{$port_name};
			} else {
				my $lock_file = $main::stat_lock_file;
				my $all_lock_files = $main::stat_dir . "/get_port.*.lock";
				my @lock_files;
				for (1..200) {
					@lock_files = glob($all_lock_files);
					last unless @lock_files;
					sleep(0.1);
				}
				if (@lock_files = glob($all_lock_files)) {
					s_log(LOG_WARNING, "Stat lock files endlessly exist, can't wait more, ignore them");
				} else {
					open(LOCKFILE, "> $lock_file") or log_n_die("Can't create lock file $lock_file");
					my $flocked = 0;
					for (1..200) {
						flock(LOCKFILE, 2) and do {
							$flocked = 1;
							last;
						};
						sleep(0.1);
					}
					unless ($flocked) {
						log_n_die("Can't lock exclusive $lock_file");
					}
				}

				my $port = get_free_port($proto, $port_base, $port_name);
				$macro = $port;
				$main::generated_ports{$port_name} = $port;

				if ($main::stat_file) {
					open(STATFILE, ">> $main::stat_file") 
						or log_n_die("Can't open stat file $main::stat_file.");
					print STATFILE "w:" . $port_name . ":" . $port . "\n" 
						or s_log(LOG_WARNING, "Can't write port info into $main::stat_file");
					close(STATFILE);
				}
				unless (@lock_files) {
					close(LOCKFILE);
					unlink($lock_file);
				}
			}
		}

		if ($macro =~ /^find_port:/) {
			(undef, my $port_name) = split(':', $macro);
			$macro = find_port_by_name($port_name);
		}

		if ($macro =~ /^remote-src:/) {
			(undef, my $host, my $port_name) = split(':', $macro);
			$macro = find_remote_port($host, $port_name);
			unless ($macro) {
				$macro_string = '';
				last;
			}
			$macro = $host . ":" . $macro;
		}

		if ($macro eq "obj-name") {
			$macro = $main::obj_name;
		}

		if ($macro eq "obj-id") {
			$macro = $main::obj_id;
		}

		if ($macro =~ /^eval:/) {
			my $eval_expression = (split(':', $macro, 2))[1];
			my $eval_result = eval($eval_expression);
			defined $eval_result or log_n_die("Error evaluation expression $eval_expression: $@");
			$macro = $eval_result;
		}

		if ($macro eq "hostname") {
			$macro = hostname;
		}


		$macro_string =~ s/(.*?)\([^\(]*?\)(.*)/$1$macro$2/;
	}
	push (@unmacro_strings, $macro_string);
}
$#unmacro_strings ? return @unmacro_strings : return $unmacro_strings[0];
}




sub cleance_config {
	my @cleared_config;
	foreach (@_) {
		push (@cleared_config, $_) if $_ && ! /^[#;]/;
	}
	return @cleared_config;
}

sub get_cfg_section {
	my $section_name = shift @_;
	my @cfg_body = @_;
	my $i;

	shift @cfg_body until (not @cfg_body or $cfg_body[0] =~ /^\[$section_name\]/);
	shift @cfg_body;

	$i++ until ($i > $#cfg_body or $cfg_body[$i] =~ /^\[/);
	$#cfg_body=--$i;
	$#cfg_body-- until ($cfg_body[-1] or not @cfg_body);
	return @cfg_body;
}


sub delete_cfg_section {
	my $section_name = shift @_;
	my $section_detected;
	my @cleared_config;

	foreach (@_) {
		if ($_ =~ /^\[$section_name\]/) {
			$section_detected = 1;
			next;
		}

		$section_detected = 0 if $section_detected && $_ =~ /^\[/;

		$section_detected or push (@cleared_config, $_);
	}
	return @cleared_config;
}


sub get_cfg_section_as_string { 
	my $some_section = join('', get_cfg_section(@_));
	$some_section =~ tr/ \t\n//d;
	return $some_section;
}
	
sub get_option {
	my $o_name = shift;
	my $o_value = (grep(/^$o_name=/, @_))[0];
	$o_value =~ s/.*?=//;

	return $o_value;
}

sub do_telnet_exchange {
	my $telnet_port = shift;
	my $telnet_password = shift;
	my @telnet_dialog = @_;
	my $vlc_daemon = new Net::Telnet (	Timeout => 2,
						errmode => "return",
						);
	my $telnet_status = $vlc_daemon->open( Host => 'localhost', Port => $telnet_port );
	if (not $telnet_status) {
		s_log(LOG_WARNING, "Error open telnet connection to localhost: $telnet_port");
		return;
	}
	$vlc_daemon->print($telnet_password);
	sleep(0.1);	
	foreach my $telnet_line (@telnet_dialog) {
		sleep(0.15) if $telnet_line =~ / play$/;
		$vlc_daemon->print($telnet_line);
		s_log(LOG_DEBUG, "$telnet_line\n");
		sleep(0.15) if $telnet_line =~ / stop$/;
		sleep(0.5) if $telnet_line =~ /^del /;
		sleep(0.05);
	}
	s_log(LOG_DEBUG, "========\n");
	$vlc_daemon->print("quit");
	sleep(0.1);
	$vlc_daemon->close;
}

1;
