package Vargus::Objects;

#    Copyright (C) 2010-2012 Michael A. Kangin <mak@complife.ru>
#
#    Copyright: Vargus is under GNU GPL, the GNU General Public License.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; version 2 dated June, 1991.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    http://www.gnu.org/licenses/gpl-2.0.html
#

use strict;
#use warnings;

use Sys::Syslog qw(openlog syslog closelog LOG_DAEMON LOG_ERR LOG_WARNING LOG_NOTICE LOG_INFO LOG_DEBUG) ;
use Vargus::Common;

use base "Exporter";
our @EXPORT = qw(
	init_objects
	get_vlm_from_objects
	);


sub get_device_by_bus {
	my $bus_id = shift;
	my $device;

	if (! %global::v4l_bus_assigned) {
		my @v4l_devices = glob("/dev/video*")
			or s_log(LOG_WARNING, "No video devices found") and return;
		foreach $device (@v4l_devices) {
			my $bus = `v4l-info $device 2>/dev/null |tr -d '[[:blank:]]' |grep '^bus_info:'`
				or s_log(LOG_WARNING, "No bus_info for device $device found") and next;
			chomp($bus);
			(undef, $bus) = split(':', $bus, 2);
			$bus =~ s/"//g 
				or s_log(LOG_WARNING, "Error bus_info for device $device processing") and next;
			$global::v4l_bus_assigned{$bus} = $device;
		}
	}
	
	$device = $global::v4l_bus_assigned{$bus_id} or s_log(LOG_WARNING, "Device not found for bus $bus_id");
	return $device;
}


sub processing_source {
	my $src = shift;
	my $origin_src = shift;
	my @cfg_body = @_;
	my $input_url;

	$input_url = $origin_src if $src eq "origin";
	$input_url = $src if $src =~ /:\/\//;
	if ( !$input_url && $src =~ /^stream.*\/target/) {
		(my $stream_name, my $target_name) = split('/', $src);
		my @stream_section = get_cfg_section($stream_name, @cfg_body);
		my $stream_target = get_option($target_name, @stream_section) or return;
		(my $target_type, my $target_url) = split(';', $stream_target, 2);
		$target_type = 'http' if $target_type eq 'mjpeg';
		$target_url = "(hostname)$target_url" if $target_url =~ /^:/;
		$target_url = "\@$target_url" if $target_type eq 'udp';

		$input_url = "$target_type://$target_url";
		$input_url =~ s/\(get_port:/(find_port:/;
	}
	return $input_url;
}

sub register_reader {
	my $input_url = shift;
	$input_url =~ /\(find_port:/ or return;

	my $port_codename = $input_url;
	$port_codename =~ s/.*\(find_port://;
	$port_codename =~ s/:.*//;
	$port_codename = expand_macroses($port_codename);

	my $port = find_port_by_name($port_codename) or log_n_die("Can't get source port for $port_codename");
	
	if ($main::stat_file) {
		open(STATFILE, ">> $main::stat_file")
			or s_log(LOG_WARNING, "Can't open stat file $main::stat_file.");
		print STATFILE "r:" . $port_codename . ":" . $port . "\n"
			or s_log(LOG_WARNING, "Can't write port info into $main::stat_file");
		close(STATFILE);
	} else {
		s_log(LOG_WARNING, "No stat file available");
	}
}




sub init_camera_objects {
	my @camera_objects = ();

	my @object_configs = sort {
		($a =~ /(\d+)/)[0] <=> ($b =~ /(\d+)/)[0] ||
		uc($a)  cmp  uc($b)
	} glob($main::conf_dir . "/*.camera.obj.cfg");

	if (not @object_configs) {
		s_log(LOG_NOTICE, "No camera object configs found");
		return;
	}

	foreach my $object_config (@object_configs) {
		open(O_CFG, $object_config) or log_n_die("Error reading object config $object_config");
		chomp(my @cfg_body = <O_CFG>);
		close(O_CFG);

		@cfg_body=cleance_config(@cfg_body);
		my %object = ();


		my @main_options = expand_macroses(get_cfg_section("main", @cfg_body));

		$object{object_type} = get_option("type", @main_options)
			or log_n_die("No 'type' in object $object_config");

		$object{object_type} eq 'camera' or log_n_die("Not 'camera' object in $object_config");

		my $is_enabled = get_option("enable", @main_options);
		if ( !$is_enabled ) {
			s_log(LOG_WARNING, "Object $object_config is not enabled, skip");
			next;
		}

		$object{active} = get_option("active", @main_options);

		$main::obj_id = $object{id} = $#{$main::vargus_objects{camera}} + 2;
		$object{config} = $object_config;

		$main::obj_name = $object{name} = get_option("name", @main_options)
			or log_n_die("No 'name' in object $object_config");

		$object{description} = get_option("description", @main_options);

		$object{write_ok} = get_option("write", @main_options);
		$object{view_ok} = get_option("view", @main_options);
		$object{farview_ok} = get_option("view", @main_options);

		
		my @import_section = get_cfg_section("import", @cfg_body);
		
		$object{import}{type} = get_option("type", @import_section)
			or $object{import}{type} = "rtsp";

		$object{import}{source} = get_option("source", @import_section)
			or log_n_die("No 'source' in object $object_config");

		if ($object{import}{type} eq "v4l2" && $object{import}{source} =~ /^BUS:/) {
			$object{import}{source} =~ s/^BUS://;
			$object{import}{source} = get_device_by_bus($object{import}{source})
				or s_log(LOG_WARNING, "Skip $object_config") and next;
		}

		$object{import}{source} = "$object{import}{type}://$object{import}{source}";

		my @vlc_options = ();
		my %known_vlc_options = ();

		$known_vlc_options{rtsp} = [('user', 'pwd', 'caching')];
		$known_vlc_options{http} = [('caching')];
		$known_vlc_options{v4l2} = [(
				'brightness', 
				'contrast', 
				'caching', 
				'height', 
				'width', 
				'fps', 
				'input', 
				'io', 
				'standard',
				'power-line-frequency'
			)];


		my $try_option;
		foreach (@{$known_vlc_options{$object{import}{type}}}) {
			$try_option = get_option($_, @import_section) 
				or $try_option = get_option("$object{import}{type}-$_", @import_section);
			push(@vlc_options, "$object{import}{type}-$_=$try_option") if $try_option;
		}

		my $raw_options;
		if ($raw_options = get_option('raw-options', @import_section)) {
			foreach(split(';', $raw_options)) {
				push(@vlc_options, $_);
			}
		}

		$object{import}{options} = [@vlc_options];

		$object{import}{loop} = get_option("loop", @import_section);
		$object{import}{aggressive} = get_option("aggressive", @import_section);


		my $stream_counter = 0;
		my @streams = ();
		my %stream_object;
		until (not %stream_object and $stream_counter > 2) {
			%stream_object = ();
			my $stream_name = "stream";
			$stream_name .= $stream_counter if $stream_counter;
			$stream_counter++;
			my @stream_section = get_cfg_section($stream_name, @cfg_body) or next;
			$stream_object{transcode} = get_option("transcode", @stream_section);

			my $target_counter = 0;
			my $stream_target = 0;
			my @targets = ();
			until (not $stream_target and $target_counter > 2) {
				my $target_name = "target";
				$target_name .= $target_counter if $target_counter;
				$target_counter++;
				$stream_target = get_option($target_name, @stream_section)
					or next;
				(my $target_type, my $target_uri) = split(';', $stream_target, 2);
				$target_type && $target_uri or log_n_die("Incorrect target: $stream_target in $object_config");
				if ($target_type eq 'rtsp') {
					$stream_target = "rtp{sdp=rtsp://$target_uri, caching=0}";
				} elsif ($target_type eq 'http') {
					$stream_target = "std{access=http,mux=ts,dst=$target_uri}";
				} elsif ($target_type eq 'mjpeg') {
					$stream_target = "std{access=http{mime=multipart/x-mixed-replace;boundary=--7b3cc56e5f51db803f790dad720ed50a},mux=mpjpeg,dst=$target_uri}";
				} elsif ($target_type eq 'udp') {
					$stream_target = "std{access=udp{caching=0},mux=ts,dst=$target_uri}";
				}

				push(@targets, $stream_target);
			}

			$stream_object{target} = [@targets];

			push(@streams, {%stream_object});
		}

		$object{stream} = [@streams];


		my @write_section = get_cfg_section("write", @cfg_body);
		$object{write}{transcode} = get_option("transcode", @write_section);
		$object{write}{path} = get_option("path", @write_section);
		$object{write}{filename} = get_option("filename", @write_section);
		$object{write}{length} = get_option("length", @write_section);
		$object{write}{start} = get_option("start", @write_section);
		$object{write}{overlap} = get_option("overlap", @write_section);
		$object{write}{random_init_delay} = get_option("random-init-delay", @write_section);
		$object{write}{source1} = processing_source(get_option("source1", @write_section), 
			$object{import}{source}, @cfg_body);
		$object{write}{source2} = processing_source(get_option("source2", @write_section),
			$object{import}{source}, @cfg_body);
		$object{write}{postprocess} = expand_macroses(get_option("postprocess", @write_section));
		$object{write}{vitality} = expand_macroses(get_option("vitality", @write_section))
			or $object{write}{vitality} = 1;

		my @view_section = get_cfg_section("view", @cfg_body);
		$object{view}{farview} = processing_source(get_option("farview", @view_section), 
			$object{import}{source}, @cfg_body) if $object{farview_ok};

		$object{view}{source} = processing_source(get_option("source", @view_section), 
			$object{import}{source}, @cfg_body) if $object{view_ok};

		$object{view}{preview} = processing_source(get_option("preview", @view_section), 
			$object{import}{source}, @cfg_body) if $object{view_ok};


		my @events_section = get_cfg_section("events", @cfg_body);
		$object{events}{agent} = get_option("agent", @events_section);
		$object{events}{encoding} = get_option("encoding", @events_section);
		$object{events}{type} = get_option("type", @events_section);
		

		my @alerts_section = get_cfg_section("alerts", @cfg_body);
		$object{alerts}{enable} = get_option("enable", @alerts_section);
		$object{alerts}{vitality} = get_option("vitality", @alerts_section) 
			or $object{alerts}{vitality} = $object{write}{vitality};
		$object{alerts}{before} = get_option("before", @alerts_section);
		$object{alerts}{after} = get_option("after", @alerts_section);

		$object{alerts}{postprocess} = expand_macroses(get_option("postprocess", @alerts_section));
		$object{alerts}{postprocess} or $object{alerts}{postprocess} = $object{write}{postprocess};

		push(@camera_objects, {%object});
	}		
	return [@camera_objects];
}
		



sub init_view_objects {
	my @view_objects = ();

	my @object_configs = glob($main::conf_dir . "/*.view.obj.cfg");
	if (not @object_configs) {
		s_log(LOG_NOTICE, "No view object configs found");
		return;
	}

	foreach my $object_config (@object_configs) {
		open(O_CFG, $object_config) or log_n_die("Error reading object config $object_config");
		chomp(my @cfg_body = <O_CFG>);
		close(O_CFG);

		@cfg_body=cleance_config(@cfg_body);

		my %object = ();


		my @main_options = expand_macroses(get_cfg_section("main", @cfg_body));

		$object{object_type} = get_option("type", @main_options)
			or log_n_die("No 'type' in object $object_config");

		$object{object_type} eq 'view' or log_n_die("Not 'view' object in $object_config");

		my $is_enabled = get_option("enable", @main_options);
		if ( !$is_enabled ) {
			s_log(LOG_WARNING, "Object $object_config is not enabled, skip");
			next;
		}


		$main::obj_id = $object{id} = $#{$main::vargus_objects{view}} + 2;
		$object{config} = $object_config;

		$main::obj_name = $object{name} = get_option("name", @main_options)
			or log_n_die("No 'name' in object $object_config");

		$object{description} = get_option("description", @main_options);

		
		my @geometry_section = get_cfg_section("geometry", @cfg_body);
		$object{geometry}{width} = get_option("width", @geometry_section)
			or log_n_die("No 'width' parameter in $object_config");
		$object{geometry}{height} = get_option("height", @geometry_section)
			or log_n_die("No 'height' parameter in $object_config");

		$object{geometry}{double} = get_option("double", @geometry_section);
		$object{geometry}{triple} = get_option("triple", @geometry_section);
		$object{geometry}{quadruple} = get_option("quadruple", @geometry_section);
		
		push(@view_objects, {%object});
	}		
	return [@view_objects];
}


sub init_set_objects {
	my @set_objects = ();

	my @object_configs = glob($main::conf_dir . "/*.set.obj.cfg");
	if (not @object_configs) {
		s_log(LOG_NOTICE, "No set object configs found");
		return;
	}

	my %actual_cameras;
	foreach(@{$main::vargus_objects{camera}}) {
		my $cam_name = ${$_}{name};
		$actual_cameras{$cam_name} = 1 if $cam_name;
	}

	foreach my $object_config (@object_configs) {
		open(O_CFG, $object_config) or log_n_die("Error reading object config $object_config");
		chomp(my @cfg_body = <O_CFG>);
		close(O_CFG);

		@cfg_body=cleance_config(@cfg_body);

		my %object = ();


		my @main_options = expand_macroses(get_cfg_section("main", @cfg_body));

		$object{object_type} = get_option("type", @main_options)
			or log_n_die("No 'type' in object $object_config");

		$object{object_type} eq 'set' or log_n_die("Not 'set' object in $object_config");

		my $is_enabled = get_option("enable", @main_options);
		if ( !$is_enabled ) {
			s_log(LOG_WARNING, "Object $object_config is not enabled, skip");
			next;
		}


		$main::obj_id = $object{id} = $#{$main::vargus_objects{set}} + 2;
		$object{config} = $object_config;

		$main::obj_name = $object{name} = get_option("name", @main_options)
			or log_n_die("No 'name' in object $object_config");

		$object{description} = get_option("description", @main_options);

		$object{cameras} = get_option("cameras", @main_options)
			or log_n_die("No 'cameras' parameter in $object_config");

		my @set_actual_cameras = ();
		
		foreach(split(',', $object{cameras})) {
			push(@set_actual_cameras, $_) if $actual_cameras{$_};
		}
		$object{cameras} = join(',', @set_actual_cameras);
			
		push(@set_objects, {%object});
	}		
	return [@set_objects];
}



sub init_objects {
	%main::vargus_objects = ();
	$main::vargus_objects{camera} = init_camera_objects;
	$main::vargus_objects{view} = init_view_objects;
	$main::vargus_objects{set} = init_set_objects;

	$main::has_objects = 1 if %main::vargus_objects;
}

sub get_vlm_from_objects {
	foreach (@{$main::vargus_objects{camera}}) {
		my %object = %{$_};
		my %vlm_desc = ();
		my $output = '';

		unless ($object{active}) {
			s_log(LOG_INFO, "Object $object{config} is not active, skip");
			next;
		}

		$vlm_desc{chan_num} = $#main::vlm_channels + 2;
		$vlm_desc{config} = $object{config};
		$vlm_desc{name} = "obj_$object{name}_$vlm_desc{chan_num}";
		$main::obj_name = $vlm_desc{obj_name} = $object{name};
		$main::obj_id = $object{id};
		$vlm_desc{variant} = 0;

		if ($main::daemon_role eq 'player') {
			$vlm_desc{loop} = $object{import}{loop};
			$vlm_desc{vlc_options} = $object{import}{options};
			$vlm_desc{input}[0] = $object{import}{source};
			$vlm_desc{aggressive} = $object{import}{aggressive};

			if ($#{$object{stream}}) {
				#More there one streams detected
				$output = 'duplicate{' . "dst='#'," x ($#{$object{stream}} + 1) . '}';
			} else {
				$output = '#';
			}

			foreach (@{$object{stream}}) {
				my %stream_object = %{$_};
				my $stream = '';
				$stream = $stream . 'transcode{' . expand_macroses($stream_object{transcode}) . '}:' 
						if $stream_object{transcode};
				$stream = $stream . 'gather:';

				if ($#{$stream_object{target}}) {
					$stream = $stream . 'duplicate{' . 'dst="#",' x ($#{$stream_object{target}} + 1) . '}';
				} else {
					$stream = $stream . '#';
				}

				foreach (@{$stream_object{target}}) {
					my $target = expand_macroses($_);
					$stream =~ s/#/$target/;
				}

				$output =~ s/#/$stream/;
			}

			$vlm_desc{output} = '#' . $output;
			$vlm_desc{output} =~ /['"]/ or $vlm_desc{output} = "'$vlm_desc{output}'";

		} elsif ($main::daemon_role eq 'writer') {
			if (!$object{write_ok}) {
				s_log(LOG_WARNING, "Write for object $object{config} is not enabled, skip");
				next;
			}
			$vlm_desc{repeat_mode} = 1;
			$vlm_desc{repeat_interval} = $object{write}{length};
			$vlm_desc{countdown} = $vlm_desc{chan_num};
			$vlm_desc{countdown} = int(rand(10)) + 1 if $object{write}{random_init_delay};
			$vlm_desc{overlap} = $object{write}{overlap};
			$vlm_desc{ttl} = $vlm_desc{repeat_interval} + $vlm_desc{overlap};
			$vlm_desc{ttl} > $main::vlc_write_maxtime 
				and log_n_die("Error: in object $object{config} write length + overlap can't be longer than vlc_write_maxtime ($main::vlc_write_maxtime)");
			if ($main::common_process) {
				$vlm_desc{input}[0] = expand_macroses($object{write}{source1});
				register_reader($object{write}{source1});
				$vlm_desc{input}[1] = expand_macroses($object{write}{source2}) 
					and register_reader($object{write}{source2}) if $vlm_desc{overlap};
			} else {
				$vlm_desc{input}[0] = $object{write}{source1};
				$vlm_desc{input}[1] = $object{write}{source2};
			}
													 
			$vlm_desc{output} = "#std{access=file,dst=$main::video_storage/$object{write}{path}/$object{write}{filename}}";
			$vlm_desc{output} =~ s/^#/#transcode{$object{write}{transcode}}:/ if $object{write}{transcode};

			$vlm_desc{filename} = $object{write}{filename};
			$vlm_desc{path} = $object{write}{path};
			$vlm_desc{start} = $object{write}{start};
			$vlm_desc{postprocess} = $object{write}{postprocess};
			$vlm_desc{vitality} = $object{write}{vitality};
		}

		push(@main::vlm_channels, {%vlm_desc});
	}
}


1;

